#include "dr_api.h"
#include "dr_ir_opnd.h"
#include "dr_ir_instr.h"
#include "dr_ir_opcodes_x86.h"
#include "dr_ir_macros_x86.h"

#include "CC_QTLib.h"

#include <ctype.h>
#include <string.h>
#include <windows.h>
#include <commdlg.h>
#include <stdint.h>

char *CUSTOM_FILE = "C:\\Users\\vboxuser\\Downloads\\spoofed_number.txt";

static app_pc pc_GetOpenFileNameA = NULL;

void (*QString_const)(void *, char *) = NULL;
void (*QArray_allocate)(void *, int64_t, int64_t, int64_t, int64_t) = NULL;
static app_pc pc_QUrl_const = NULL;
static app_pc pc_QList_const = NULL;

static app_pc pc_QFileDialog_getOpenFileName = NULL;
static app_pc pc_QFileDialog_getOpenFileNames = NULL;
static app_pc pc_QFileDialog_getOpenFileUrl = NULL;
static app_pc pc_QFileDialog_getOpenFileUrls = NULL;
static app_pc pc_QFileDialog_getOpenFileContent = NULL;

static void intercept_GetOpenFileNameA() {
    void *drcontext = dr_get_current_drcontext();

    dr_mcontext_t mcontext = { sizeof(mcontext), DR_MC_ALL };
    dr_get_mcontext(drcontext, &mcontext);

    OPENFILENAME *ofn = (OPENFILENAME *)mcontext.rcx;

    for (int i = 0; i < strlen(CUSTOM_FILE); i++) ofn->lpstrFile[i] = CUSTOM_FILE[i];
    ofn->lpstrFile[strlen(CUSTOM_FILE)] = 0;

    mcontext.rax = 1;

    dr_set_mcontext(drcontext, &mcontext);
}

static void intercept_QFileDialog_getOpenFileName() {
    void *drcontext = dr_get_current_drcontext();

    dr_printf("intercept_QFileDialog_getOpenFileName\n");

    dr_mcontext_t mcontext = { sizeof(mcontext), DR_MC_ALL };
    dr_get_mcontext(drcontext, &mcontext);

    CC_QString_insert((void *)mcontext.rcx, CUSTOM_FILE);
}

static void intercept_QFileDialog_getOpenFileNames() {
    void *drcontext = dr_get_current_drcontext();

    dr_mcontext_t mcontext = { sizeof(mcontext), DR_MC_ALL };
    dr_get_mcontext(drcontext, &mcontext);

    // make use of QString constructor to construct return value
    // void *QArray = (void *)mcontext.rcx;

    // char QString[24];
    // QString_const(QString, CUSTOM_FILE);

    // QArray_allocate(QArray, 0x18, 8, 1, 1);
}


static void event_exit(void) {
}

static void event_module_load(void *drcontext, const module_data_t *mod, bool loaded) {
    char lowerpath[MAX_PATH];
    strcpy(lowerpath, mod->full_path);
    for (int i = 0; i < strlen(lowerpath); i++) {
        lowerpath[i] = tolower(lowerpath[i]);
    }

    if (strstr(lowerpath, "qt6core.dll")) {
        QString_const = (app_pc)dr_get_proc_address(mod->handle, "??0QString@@QEAA@PEBD@Z");
        QArray_allocate = (app_pc)dr_get_proc_address(mod->handle, "?allocate@QArrayData@@SAPEAXPEAPEAU1@_J11W4AllocationOption@1@@Z");
        // pc_QStringList_const = (app_pc)dr_get_proc_address(mod->handle, "??0QString@@QEAA@PEBD@Z");
        // if (pc_QString_const) dr_printf("Found QString::QString at %p\n", pc_QString_const);
    }
    if (strstr(lowerpath, "qt6widgets.dll")) {
        pc_QFileDialog_getOpenFileName = (app_pc)dr_get_proc_address(mod->handle, "?getOpenFileName@QFileDialog@@SA?AVQString@@PEAVQWidget@@AEBV2@11PEAV2@V?$QFlags@W4Option@QFileDialog@@@@@Z");
        pc_QFileDialog_getOpenFileNames = (app_pc)dr_get_proc_address(mod->handle, "?getOpenFileNames@QFileDialog@@SA?AV?$QList@VQString@@@@PEAVQWidget@@AEBVQString@@11PEAV4@V?$QFlags@W4Option@QFileDialog@@@@@Z");
        pc_QFileDialog_getOpenFileUrl = (app_pc)dr_get_proc_address(mod->handle, "?getOpenFileUrl@QFileDialog@@SA?AVQUrl@@PEAVQWidget@@AEBVQString@@AEBV2@1PEAV4@V?$QFlags@W4Option@QFileDialog@@@@AEBV?$QList@VQString@@@@@Z");
        pc_QFileDialog_getOpenFileUrls = (app_pc)dr_get_proc_address(mod->handle, "?getOpenFileUrls@QFileDialog@@SA?AV?$QList@VQUrl@@@@PEAVQWidget@@AEBVQString@@AEBVQUrl@@1PEAV4@V?$QFlags@W4Option@QFileDialog@@@@AEBV?$QList@VQString@@@@@Z");
        pc_QFileDialog_getOpenFileContent = (app_pc)dr_get_proc_address(mod->handle, "?getOpenFileContent@QFileDialog@@SAXAEBVQString@@AEBV?$function@$$A6AXAEBVQString@@AEBVQByteArray@@@Z@std@@PEAVQWidget@@@Z");
        // if (pc_QFileDialog_getOpenFileName) dr_printf("Found QFileDialog::getOpenFileName at %p\n", pc_QFileDialog_getOpenFileName);
    }

    if (strstr(lowerpath, "comdlg32.dll")) {
        pc_GetOpenFileNameA = (app_pc)dr_get_proc_address(mod->handle, "GetOpenFileNameA");
        // if (pc_GetOpenFileNameA) dr_printf("Found GetOpenFileNameA at %p\n", pc_GetOpenFileNameA);
    }
}


static dr_emit_flags_t event_basic_block(void *drcontext, void *tag, instrlist_t *bb, bool for_trace, bool translating) {
    instr_t *to_remove = NULL, *instr = instrlist_first(bb);
    while (instr) {
        instr_t *next = instr_get_next(instr);

        app_pc target = NULL;

        if (instr_is_call_direct(instr)) {
            app_pc target = instr_get_branch_target_pc(instr);
        } else if (instr_is_call_indirect(instr)) {
            opnd_t opnd = instr_get_src(instr, 0);
            if (opnd_is_memory_reference(opnd)) {
                app_pc addr = opnd_compute_address(opnd, drcontext);
                if (addr) dr_safe_read(addr, sizeof(app_pc), &target, NULL);
            }
        }

        if (!target) {
            instr = next;
            continue;
        }

        if (target == pc_GetOpenFileNameA) {
            // dr_printf("Call to GetOpenFileNameA detected at %p\n", instr_get_app_pc(instr));
            dr_insert_clean_call(drcontext, bb, instr, (void *)intercept_GetOpenFileNameA, false, 0);
            instrlist_remove(bb, instr);
            instr_destroy(drcontext, instr);
        }
        else if (target == pc_QFileDialog_getOpenFileName) {
            // dr_printf("Call to QFileDialog::getOpenFileName detected at %p\n", instr_get_app_pc(instr));
            dr_insert_clean_call(drcontext, bb, instr, (void *)intercept_QFileDialog_getOpenFileName, false, 0);
            instrlist_remove(bb, instr);
            instr_destroy(drcontext, instr);
        }

        instr = next;
    }

    return DR_EMIT_DEFAULT;
}

DR_EXPORT void dr_client_main(client_id_t id, int argc, const char *argv[]) {
    dr_set_client_name("Custom DynamoRIO client", "https://gitea.cloud.lehnert.dev/ludwig/windows-binary-fuzzing");
    dr_register_exit_event(event_exit);
    dr_register_module_load_event(event_module_load);
    dr_register_bb_event(event_basic_block);

    // CC_QTLib_load();
}
